#!/usr/bin/ruby
# encoding: utf-8

require 'glib2'
require 'gtk2'
require 'vte'
require "rexml/document"
require 'MfsDiskDrive'
require 'MfsSinglePartition'

#### Global variables

@drivelist = Array.new

#### Commonly used functions

def umount_stick(device)
	device.partitions.each { |p| 
		p.umount
	}
	return device.mounted 
end

def fill_drive_combo(dropdown, apply)
	99.downto(0) { |n|
		begin
			dropdown.remove_text(n)
		rescue
		end
	}
	@drivelist = Array.new
	drives = Array.new
	Dir.entries("/sys/block").each { |l|
		drives.push(MfsDiskDrive.new(l, true)) if l =~ /[a-z]$/ 
	}
	drives.each { |d| 
		if d.usb == true && d.size > 3_700_000_000 && ( d.mounted == false || umount_stick(d) == false )
			@drivelist.push d
			dropdown.append_text "#{d.vendor} #{d.model} #{d.device} #{d.human_size}"
		end
	} 
	if @drivelist.size > 0
		apply.sensitive = true
		dropdown.sensitive = true
	else
		apply.sensitive = false
		dropdown.sensitive = false
		dropdown.append_text "Kein geeignetes Laufwerk gefunden"
	end
	dropdown.active = 0
	
end

#### Modal dialog to confirm

def confirm_dialog(drive)
	dialog = Gtk::Dialog.new(
		"Wollen Sie wirklich installieren?",
		$mainwindow,
		Gtk::Dialog::MODAL,
		[ Gtk::Stock::YES, Gtk::Dialog::RESPONSE_YES ],
		[ Gtk::Stock::NO, Gtk::Dialog::RESPONSE_NO ]
	)
	dialog.default_response = Gtk::Dialog::RESPONSE_NO
	dialog.has_separator = false
	label = Gtk::Label.new
	label.set_markup("Wollen Sie Desinfec't wirklich auf dem USB-Laufwerk #{drive.model} #{drive.vendor} #{drive.human_size} installieren?\n\n<b>Hierbei werden alle auf dem Laufwerk gespeicherten Daten gelöscht!</b>")
	label.wrap = true
	image = Gtk::Image.new(Gtk::Stock::DIALOG_WARNING, Gtk::IconSize::DIALOG)
	hbox = Gtk::HBox.new(false, 5)
	hbox.border_width = 10
	hbox.pack_start_defaults(image)
	hbox.pack_start_defaults(label)
	dialog.vbox.add(hbox)
	dialog.show_all
	dialog.run { |resp|
		case resp
		when Gtk::Dialog::RESPONSE_YES
			dialog.destroy
			return true
		else
			dialog.destroy
			return false
		end
	}
end

def confirm_again_dialog(drive)
	dialog = Gtk::Dialog.new(
		"Wollen Sie wirklich installieren?",
		$mainwindow,
		Gtk::Dialog::MODAL,
		[ Gtk::Stock::YES, Gtk::Dialog::RESPONSE_YES ],
		[ Gtk::Stock::NO, Gtk::Dialog::RESPONSE_NO ]
	)
	dialog.default_response = Gtk::Dialog::RESPONSE_NO
	dialog.has_separator = false
	label = Gtk::Label.new
	label.set_markup("<span size='large' weight='bold'>ACHTUNG!</span>\n\nDer ausgewählte Datenträger #{drive.vendor} #{drive.model}  ist\n\n<span size='large' weight='bold'>#{drive.human_size}</span>\n\ngroß und soll komplett gelöscht werden. Sind Sie wirklich ganz sicher, dass es sich nicht um eine Festplatte handelt, die Sie noch brauchen?\n\nKlicken Sie auf \"Ja\", um das Laufwerk zu löschen und Desinfec't zu installieren, mit \"Nein\" gelangen Sie zurück zur Laufwerksauswahl.")
	label.wrap = true
	image = Gtk::Image.new(Gtk::Stock::DIALOG_WARNING, Gtk::IconSize::DIALOG)
	hbox = Gtk::HBox.new(false, 5)
	hbox.border_width = 10
	# hbox.pack_start_defaults(image)
	hbox.pack_start_defaults(label)
	dialog.vbox.add(hbox)
	dialog.show_all
	dialog.run { |resp|
		case resp
		when Gtk::Dialog::RESPONSE_YES
			dialog.destroy
			return true
		else
			dialog.destroy
			return false
		end
	}
end

#### Update progress bar

def update_pgbar(pgbar, text) 
	while (Gtk.events_pending?)
		Gtk.main_iteration
	end
	pgbar.text = text
	pgbar.pulse
	while (Gtk.events_pending?)
		Gtk.main_iteration
	end
end

#### Modal dialog for installation progress

def install_progress(drive, gptprefered=true, easymode=false)
	dialog = Gtk::Dialog.new(
		"Installationsfortschritt",
		$mainwindow,
		Gtk::Dialog::MODAL
	)
	dialog.deletable = false
	label = Gtk::Label.new
	label.width_request = 400
	label.set_markup("Die Installation wird durchgeführt, bitte ziehen Sie den USB-Stick während der Installation nicht ab.")
	label.wrap = true
	image = Gtk::Image.new(Gtk::Stock::DIALOG_INFO, Gtk::IconSize::DIALOG)
	hbox = Gtk::HBox.new(false, 5)
	hbox.border_width = 10
	hbox.pack_start_defaults(image)
	hbox.pack_start_defaults(label)
	pgbar = Gtk::ProgressBar.new
	okbutton = Gtk::Button.new(Gtk::Stock::OK)
	okbutton.sensitive = false
	dialog.vbox.add(hbox)
	dialog.vbox.add(pgbar)
	dialog.add_action_widget(okbutton, Gtk::Dialog::RESPONSE_OK)
	dialog.show_all
	unpluggable = run_installation(drive, pgbar, gptprefered, easymode)
	if unpluggable == true
		pgbar.text = "Installation abgeschlossen"
		label.set_markup("Die Installation ist abgeschlossen, Sie können den USB-Stick jetzt abziehen.")
	else
		pgbar.text = "Installation abgeschlossen"
		label.set_markup("Die Installation ist abgeschlossen, der USB-Stick wird nun verwendet und darf nicht abgezogen werden.")
	end
	pgbar.fraction = 1.0 
	okbutton.sensitive = true
	dialog.run
	dialog.destroy
end

#### Run the real installation 

def run_installation(target, pgbar, gptprefered=true, easymode=false)
	# some calculations
	sigsize = 1280
	swapsize = 256
	if target.size > 15_000_000_000
		sigsize = 8192
		swapsize = 1535
	elsif target.size > 7_000_000_000
		sigsize = 4096
		swapsize = 1023
	end
	uuid = nil
	puts target.size.to_s
	kblocks = target.size / 1024
	puts kblocks.to_s
	update_pgbar(pgbar, "Lösche USB-Stick...")
	# Zap partition contents
	4.downto(1) { |n|
		command = "dd if=/dev/zero bs=1024 count=1024 of=/dev/#{target.device}#{n.to_s}"
		puts command
		system command
	}
	# Zap GPT / MBR
	command = "dd if=/dev/zero bs=1024 count=1024 of=/dev/#{target.device}"
	puts command
	system command
	command = "dd if=/dev/zero bs=1024 count=2048 of=/dev/#{target.device} seek=#{kblocks - 2048}"
	puts command 
	system command 
	# Create partitions
	pend = Array.new
	# Fixed size for partition 4 - UEFI boot
	pend[3] = target.size - 1048577
	# Fixed size for partition 3 - signatures, swapfile, persistent data
	pend[2] = pend[3] - ( 64 * 1048576)
	# Fixed size for partition 2 - system and BIOS boot
	pend[1] = pend[2] - ( sigsize * 1048576)
	# Flexible size for partition 1 - user data
	pend[0] = pend[1] - ( 1792 * 1048576)
	update_pgbar(pgbar, "Erstelle Partitionstabelle...")
	if gptprefered == true
		command = "/usr/local/parted-3.1/sbin/parted -s /dev/#{target.device} unit B mklabel gpt"
	else
		command = "/usr/local/parted-3.1/sbin/parted -s /dev/#{target.device} unit B mklabel msdos"
	end
	puts command
	system command
	update_pgbar(pgbar, "Erstelle Partition 1...")
	command = "/usr/local/parted-3.1/sbin/parted -s /dev/#{target.device} unit B mkpart primary fat32 1048576 #{pend[0].to_s} "
	puts command 
	system command
	update_pgbar(pgbar, "Erstelle Partition 2...")
	command = "/usr/local/parted-3.1/sbin/parted -s /dev/#{target.device} unit B mkpart primary fat32 #{ (pend[0] + 1).to_s} #{pend[1].to_s} "
	puts command
	system command
	update_pgbar(pgbar, "Erstelle Partition 3...")
	command = "/usr/local/parted-3.1/sbin/parted -s /dev/#{target.device} unit B mkpart primary ext3 #{ (pend[1] + 1).to_s} #{pend[2].to_s} "
	puts command 
	system command
	update_pgbar(pgbar, "Erstelle Partition 4...")
	command = "/usr/local/parted-3.1/sbin/parted -s /dev/#{target.device} unit B mkpart primary fat32 #{ (pend[2] + 1).to_s} #{pend[3].to_s} "
	puts command
	system command
	sleep 5
	
	# Mount desinfDATA - is needed later!
	unless system("mountpoint -q /media/desinfDATA")
		if system("test -b /dev/disk/by-label/desinfDATA")
			system("mkdir -p /media/desinfDATA")
			system("mount -t vfat -o uid=rw,999,gid=999 /dev/disk/by-label/desinfDATA /media/desinfDATA") 
		end
	end
	
	1.upto(4) { |n| system("umount -f /dev/#{target.device}#{n}") }
	update_pgbar(pgbar, "Setze Partitionseigenschaften...")
	if gptprefered == true
		# Set boot flag UEFI
		# Set legacy boot flag BIOS
		command = "/usr/local/parted-3.1/sbin/parted -s /dev/#{target.device} unit B set 4 boot on"
		puts command
		system command
		command = "/usr/local/parted-3.1/sbin/parted -s /dev/#{target.device} unit B set 2 legacy_boot on"
		puts command
		system command
		# Install legacy MBR
		command = "dd if=/usr/lib/syslinux/gptmbr.bin bs=440 count=1 of=/dev/#{target.device}"
		puts command
		system command
	else
		command = "/usr/local/parted-3.1/sbin/parted -s /dev/#{target.device} unit B set 2 boot on"
		puts command
		system command
		# Install legacy MBR
		command = "dd if=/usr/lib/syslinux/mbr.bin bs=440 count=1 of=/dev/#{target.device}"
		puts command
		system command
	end
	# Make filesystems
	fscomm = [ "mkfs.msdos -F32 -n desinfDATA /dev/#{target.device}1", 
		"mkfs.msdos -F32 -n desinfSYS /dev/#{target.device}2", 
		"mkfs.ext4 -L desinfSIGS /dev/#{target.device}3" ]
	fscomm.push("mkfs.msdos -F32 /dev/#{target.device}4") if gptprefered == true
	fscomm.each { |m|
		update_pgbar(pgbar, "Erstelle Dateisysteme...")
		puts m
		unless system(m)
			sleep 2
			unless system(m)
				sleep 5
				unless system(m)
					puts "Installation fehlgeschlagen... Konnte #{m} nicht ausführen!"  
					exit 1
				end
			end
		end
	}
	sleep 5
	1.upto(4) { |n| system("umount -f /dev/#{target.device}#{n}") } 
	# Retrieve UUID for GRUB 
	IO.popen("blkid -o udev /dev/#{target.device}2") { |l|
		while l.gets
			toks = $_.strip.split("=")
			uuid = toks[1] if toks[0] == "ID_FS_UUID"
		end
	}
	# Install shutdown.bat
	system("mkdir -p /desinfect/target")
	system("mount -t vfat /dev/#{target.device}1 /desinfect/target")
	if File.exists?("/cdrom/shutdown.bat")
		system("cp -v /cdrom/shutdown.bat /desinfect/target")
	else
		system("cp -v /opt/desinfect/usb-installer/shutdown.bat /desinfect/target")
	end
	tvdir = nil
	if File.directory?("/cdrom/TeamViewer Portable")
		tvdir = "/cdrom/"
	elsif File.directory?("/isodevice/TeamViewer Portable")
		tvdir = "/isodevice/"
	elsif File.directory?("/media/desinfDATA/TeamViewer Portable")
		tvdir = "/media/desinfDATA/"	
	end
	system("cp -vr #{tvdir}'TeamViewer Portable' /desinfect/target/" ) unless tvdir.nil?
	win8dir = nil
	if File.directory?("/cdrom/Windows 8 Hilfe")
		win8dir = "/cdrom/"
	elsif File.directory?("/isodevice/Windows 8 Hilfe")
		win8dir = "/isodevice/"
	elsif File.directory?("/media/desinfDATA/Windows 8 Hilfe")
		win8dir = "/media/desinfDATA/"	
	end
	system("cp -vr #{win8dir}'Windows 8 Hilfe' /desinfect/target/" ) unless win8dir.nil?
	system("cp -v /cdrom/hiberfil-abschalten.bat /desinfect/target")
	system("cp -v /cdrom/hiberfil-anschalten.bat /desinfect/target")
	system("sync")
	unless system("umount /dev/#{target.device}1") 
		sleep 2
		system("umount /dev/#{target.device}1") 
	end
	if system("mountpoint /desinfect/target")
		puts "Installation fehlgeschlagen... Data-Partition konnte nicht ausgehängt werden!"  
		exit 1
	end
	# Install base system
	install_running = true
	update_pgbar(pgbar, "Installiere Basissystem...")
	unless system("mount -t vfat /dev/#{target.device}2 /desinfect/target")
		sleep 2
		system("umount -f /dev/#{target.device}2")
		system("mount -t vfat /dev/#{target.device}2 /desinfect/target")
	end
	unless system("mountpoint /desinfect/target")
		puts "Installation fehlgeschlagen... System-Partition konnte nicht eingehängt werden!"  
		exit 1
	end
	copy_term = Vte::Terminal.new
	copy_term.fork_command("bash", [ "bash", "copy.sh"] )
	copy_term.signal_connect("child-exited") { install_running = false } 
	# system("tar -C /cdrom -cf - casper boot isolinux | tar -C /desinfect/target -xvf - ") 
	while install_running == true 
		sleep 0.5
		pgbar.pulse
		while (Gtk.events_pending?)
			Gtk.main_iteration
		end
	end
	update_pgbar(pgbar, "Installiere Bootloader (BIOS)...")
	# Install syslinux 
	system("cp /desinfect/target/isolinux/isolinux.cfg /desinfect/target/isolinux/syslinux.cfg")
	if File.exist?('/desinfect/target/isolinux/os.cfg.orig')
		system("cp -v /desinfect/target/isolinux/os.cfg.orig /desinfect/target/isolinux/os.cfg")
	else
		system("cp -v /desinfect/target/isolinux/os.cfg /desinfect/target/isolinux/os.cfg.orig")
	end
	if easymode == true
		system('sed -i \'s/label desinfect-easy/label desinfect-easy\n  menu default/g\' /desinfect/target/isolinux/os.cfg')  
		# system('sed -i \'s/set default=0/set default=1/g\' /desinfect/target/boot/grub/grub.cfg') 
	end
	system("sync") 
	unless system("umount /dev/#{target.device}2") 
		sleep 2
		system("umount -f /dev/#{target.device}2") 
	end
	1.upto(4) { |n| 
		unless system("umount /dev/#{target.device}#{n}") 
			sleep 1
			system("umount -f /dev/#{target.device}#{n}") 
		end
	}  
	if system("mountpoint /desinfect/target")
		puts "Installation fehlgeschlagen... System-Partition konnte nicht ausgehängt werden!"  
		exit 1
	end
	system("syslinux --install -d /isolinux /dev/#{target.device}2")
	if gptprefered == true
		# Install UEFI boot files
		update_pgbar(pgbar, "Installiere Bootloader (UEFI)...")
		unless system("mount -t vfat /dev/#{target.device}4 /desinfect/target")
			sleep 2
			system("mount -t vfat /dev/#{target.device}4 /desinfect/target")
		end
		unless system("mountpoint /desinfect/target")
			puts "Installation fehlgeschlagen... UEFI-Partition konnte nicht eingehängt werden!"  
			exit 1
		end
		system("tar -C /cdrom -cf - EFI unused.EFI casper/initrd.lz casper/vmlinuz boot | tar -C /desinfect/target -xvf - ")
		system("mv /desinfect/target/unused.EFI /desinfect/target/EFI")
		if File.exist?('/desinfect/target/boot/grub/grub.cfg.orig')
			system("cp -v /desinfect/target/boot/grub/grub.cfg.orig /desinfect/target/boot/grub/grub.cfg")
		else
			system("cp -v /desinfect/target/boot/grub/grub.cfg /desinfect/target/boot/grub/grub.cfg.orig")
		end
		system('sed -i \'s/set default=0/set default=1/g\' /desinfect/target/boot/grub/grub.cfg') if easymode == true
		# Replace UUID in GRUB config
		# system("sed -i 's/TARGETUUID/#{uuid}/g' /desinfect/target/EFI/BOOT/grub/grub.cfg ")
		system("sync") 
		unless system("umount /dev/#{target.device}4")
			sleep 2
			system("umount /dev/#{target.device}4")
		end
		if system("mountpoint /desinfect/target")
			puts "Installation fehlgeschlagen... UEFI-Partition konnte nicht ausgehängt werden!"  
			exit 1
		end
	end
	# Create Swap file
	# Copy signatures
	update_pgbar(pgbar, "Kopiere Signaturen...")
	unless system("mount -t ext4 /dev/#{target.device}3 /desinfect/target")
		sleep 2
		system("mount -t ext4 /dev/#{target.device}3 /desinfect/target")
	end
	unless system("mountpoint /desinfect/target")
		puts "Installation fehlgeschlagen... Signatur-Partition konnte nicht eingehängt werden!"  
		exit 1
	end
	system("mkdir -p /desinfect/target/desinfect-signatures")
	system("mkdir -p /desinfect/target/deb")
	[ "kaspersky", "bitdefender", "avira", "clamav" ].each { |d|
		system("mkdir -p /desinfect/target/desinfect-signatures/#{d}")
	}
	# Let our script do the real copy job
	siginst_running = true
	sigs_term = Vte::Terminal.new
	sigs_term.fork_command("bash", [ "bash", "copy-sigs.sh"] )
	sigs_term.signal_connect("child-exited") { siginst_running = false } 
	while siginst_running == true 
		sleep 0.5
		pgbar.pulse
		while (Gtk.events_pending?)
			Gtk.main_iteration
		end
	end
	# Use swap file
	system("dd if=/dev/zero bs=1M count=1 seek=#{swapsize.to_s} of=/desinfect/target/swap.img")
	system("mkswap /desinfect/target/swap.img") 
	# Use signatures
	system("sync") 
	unless system("umount /dev/#{target.device}3") 
		sleep 2
		system("umount /dev/#{target.device}3") 
	end
	1.upto(4) { |n| system("umount /dev/#{target.device}#{n}") }
	if system("mountpoint /desinfect/target")
		puts "Installation fehlgeschlagen... Signatur-Partition konnte nicht ausgehängt werden!"  
		exit 1
	end
	# ...or unmount
	retval = false
	if system("mountpoint /opt/desinfect/signatures")
		# There already is a stick present, tell the user he can remove the new one
		1.upto(4) { |n| system("umount /dev/#{target.device}#{n}") } 
		retval = true
	else
		# There is no stick present, use our stick
		system("mount /dev/#{target.device}3 /opt/desinfect/signatures")
		system("umount /var/kl/bases_rd")
		system("umount /opt/BitDefender-scanner/var/lib/scan")
		# FIXME! Umount some some potentially mounted tmpfs directories
		{ "avira" => "/AntiVir", 
		  "clamav" =>  "/var/lib/clamav", 
		  "kaspersky" => "/var/kl/bases_rd", 
		  "bitdefender" => "/opt/BitDefender-scanner/var/lib" }.each { |k,v| 
			system("mount --bind /opt/desinfect/signatures/desinfect-signatures/#{k} #{v}")
		}
		system("mkdir -p /tmp/desinfect_logs")
		system("mkdir -p /opt/desinfect/signatures/logs")
		system("touch /opt/desinfect/signatures/.desinfect2015") 
		system("sync") 
		system("mount --bind /opt/desinfect/signatures/logs /tmp/desinfect_logs")
		freeloop = ` losetup -f `.strip
		system("losetup #{freeloop} /opt/desinfect/signatures/swap.img")
		system("swapon #{freeloop}")
		retval = false
	end
	return retval
end

#### Now create widgets

window = Gtk::Window.new
window.signal_connect("delete_event") {
	puts "delete event occurred"
	false
}

window.signal_connect("destroy") {
	puts "destroy event occurred"
	Gtk.main_quit
}

#### Layout-Tabelle

ltab = Gtk::Table.new(6, 4, false)

#### Header, fat

head = Gtk::Label.new
head.set_markup("<b>Desinfec't USB-Installation</b>")
# ltab.attach(head, 0, 4, 0, 1, Gtk::EXPAND|Gtk::FILL, 5, 5) 

#### Descriptive text

desc = Gtk::Label.new
desc.set_markup("Dieses Werkzeug installiert desinfec't bootfähig auf einen wenigstens vier Gigabyte großen USB-Stick. Alle bislang auf dem USB-Stick gespeicherten Daten werden hierbei gelöscht. Die Installation dauert je nach Geschwindigkeit des USB-Sticks 5 bis 15 Minuten.")
desc.wrap = true
desc.width_request = 480
align = Gtk::Alignment.new(0.1, 0.1, 0.9, 0.9)
align.add desc
ltab.attach(align, 0, 4, 1, 2, Gtk::EXPAND|Gtk::FILL, 5, 5) 

#### Dropdown of suitable drives

tlab = Gtk::Label.new("Ziellaufwerk:")
ltab.attach(tlab, 0, 1, 2, 3, Gtk::EXPAND|Gtk::FILL, 5, 5) 

drop = Gtk::ComboBox.new
drop.width_request = 300
ltab.attach(drop, 1, 3, 2, 3, Gtk::EXPAND|Gtk::FILL, 5, 5) 

#### Re-read button

reread = Gtk::Button.new("Neu einlesen")
ltab.attach(reread, 3, 4, 2, 3, Gtk::EXPAND|Gtk::FILL, 5, 5) 

#### Checkbox for type

clab = Gtk::Label.new("Kompatibilität:")
ltab.attach(clab, 0, 1, 3, 4, Gtk::EXPAND|Gtk::FILL, 5, 5) 
gptcheck = Gtk::CheckButton.new("Sowohl BIOS als auch UEFI unterstützen")
gptcheck.active = true if ARGV[0].to_s == "gpt"
ltab.attach(gptcheck, 1, 4, 3, 4, Gtk::EXPAND|Gtk::FILL, 5, 5) 

#### Checkbox for Easy Mode

elab = Gtk::Label.new("Easy Scan:")
ltab.attach(elab, 0, 1, 4, 5, Gtk::EXPAND|Gtk::FILL, 5, 5) 
easycheck = Gtk::CheckButton.new("Automatik-Modus als Standard verwenden")
easycheck.active = false
ltab.attach(easycheck, 1, 4, 4, 5, Gtk::EXPAND|Gtk::FILL, 5, 5) 

#### Install button

fillup = Gtk::Alignment.new(0, 0, 1, 1)
fillup.width_request = 180
instbutton = Gtk::Button.new(Gtk::Stock::APPLY)
ltab.attach(fillup, 1, 2, 5, 6, Gtk::EXPAND|Gtk::FILL, 5, 5) 
ltab.attach(instbutton, 2, 3, 5, 6, Gtk::EXPAND|Gtk::FILL, 5, 5) 

#### Cancel Button

cancel = Gtk::Button.new(Gtk::Stock::CANCEL)
ltab.attach(cancel, 3, 4, 5, 6, Gtk::EXPAND|Gtk::FILL, 5, 5) 
cancel.signal_connect("clicked") { Gtk.main_quit } 

#### Wire some functions

fill_drive_combo(drop, instbutton)
reread.signal_connect("clicked") { fill_drive_combo(drop, instbutton) } 
instbutton.signal_connect("clicked")  { 
	run_installation = confirm_dialog(@drivelist[drop.active])
	if run_installation == true && @drivelist[drop.active].size > 40_000_000_000
		run_installation = confirm_again_dialog(@drivelist[drop.active])
	end
	if run_installation == true
		install_progress(@drivelist[drop.active], gptcheck.active?, easycheck.active? )
		Gtk.main_quit
	end
} 

#### 

window.border_width = 10
window.set_title("Desinfec't USB-Installation")
window.window_position = Gtk::Window::POS_CENTER_ALWAYS
ltab.row_spacings = 10
window.add ltab
window.show_all

Gtk.main

